# Version 0.0.3 - Oct 16, 2011
##########----------##########----------##########----------##########----------
# my camera: 23 x 15.5 mm
# Olypus: 13 x 17.3 mm
# Nikon APS-C: 23.6 x 15.8 mm
# Pentax K20D 23.4 x 15.6 mm
# Canon APS-C: 22.3 x 14.9 mm
# Full frame: 36 x 24mm

#####################  THE  9  PARAMETERS  YOU  MUST  SET  ARE  THERE
                    #  ##################################################
f= 50               # mm, for example 50 mm 
Vside= 36           # mm, for example 36 mm  |  portrait / landscape  
Hside= 24           # mm, for example 24 mm  |  orientation found there
phiMin= -80         # degrees, from -90 (nadir) toward 90
phiMax= 90          # degrees, up to 90 (zenith) toward -90
Hoverlap= 0.20      # horizontal overlap, coef from 0 to 0.5 (0% to 50 %)
Voverlap= 0.20      # vertical overlap, coef from 0 up to more than 0.5 
                    #
panoheadPhiMin= -90 # degrees, from -90 (nadir)  | avoids the camera or lens
panoheadPhiMax= 90  # degrees, up to 90 (zenith) | hitting the pano head !
                    #
#####################
import os
import sys
if len(sys.argv)>1:
    f = int(sys.argv[1]) 
from math import * 

assert f > 1 and f < 1001, "f: improper focal length"
assert Vside > 1 and Vside < 100, "Vside: improper size"
assert Hside > 1 and Hside < 100, "Vside: improper size"
assert phiMin >= -90 and phiMin < 85, "phiMin not between -90 and +85"
assert phiMax <= 90 and phiMax > -85, "phiMax not between -85 and +90"
assert phiMin != phiMax, "phiMin, phiMax: min and max are equal"
assert phiMin < phiMax, "phiMin, phiMax: min and max are inverted"
assert Hoverlap >= 0, "Hoverlap: minimum possible value is 0"
assert Hoverlap < 0.96, "Hoverlap: maximum possible value is 0.95"
assert Voverlap >= 0, "Voverlap: minimum value is 0"
assert Voverlap < 0.96, "Voverlap: maximum value is 0.95"
assert panoheadPhiMin >= -90, "panoheadPhiMin: extreme value is -90"
assert panoheadPhiMax <= 90, "panoheadPhiMax: maximum value is 90"
assert panoheadPhiMin < panoheadPhiMax, "panoheadPhiMin, panoheadPhiMax:" \
       + "Pano-head extreme top & bottom positions are stupid!"
    
trueVside=  Vside
trueHside=  Hside
trueVfov=   degrees(2 * atan (trueVside / 2 / f))
trueHfov=   degrees(2 * atan (trueHside / 2 / f))

if phiMax - trueVfov / 2 > panoheadPhiMax :
    print("\nCAUTION! ...panorama phiMax incompatible", 
          "with pano-head phiMax has been reduced!\n")
    phiMax= panoheadPhiMax - trueVfov / 2 

if phiMin + trueVfov / 2 < panoheadPhiMin :
    phiMin= panoheadPhiMin + trueVfov / 2
    print("\nCAUTION! ...panorama phiMin incompatible",
          "with pano-head phiMin has been adjusted!\n")

truePhiMin= phiMin                             
truePhiMax= phiMax                             
                                                
##########----------##########----------##########----------##########----------

print("Camera:")
print("    Sensor size         ", repr(trueHside)[0:4] + " x " + repr(trueVside),"mm")
print("    Focal length        ", repr(f)[0:5], "mm")
print("    Lens vertical FOV   ", round(degrees(2 * atan (trueVside / 2 / f))), "deg")
print("    Lens horizontal FOV ", round(degrees(2 * atan (trueHside / 2 / f))), "deg")

print("\nPanorama:")
print("    Top at              ", repr(phiMax)[0:5], "deg")
print("    Bottom at           ", repr(phiMin)[0:5], "deg")
print("    vertical FOV        ", repr(phiMax - phiMin)[0:5], "deg")
print("    Vertical overlap    ", repr(int(100 * Voverlap))+" %")
print("    Horizontal overlap  ", repr(int(100 * Hoverlap))+" %")

if panoheadPhiMin != -90 or panoheadPhiMax != 90 :
    print("\nPano-head:")
    if panoheadPhiMin != -90  :
        print("    downward limit", int(panoheadPhiMin), " deg")
    if panoheadPhiMax != 90 :
        print("    upward limit", int(panoheadPhiMax), " deg")

print(" ")

##########----------##########----------##########----------##########----------

totalImgNb= int(0)

if trueVfov > trueHfov :
    optimalZenith= 90 - ((trueVfov - trueHfov) / 2)
    optimalNadir= -90 + ((trueVfov - trueHfov) / 2)
else :
    optimalZenith = 90
    optimalNadir = -90               # overlap between zenith disc and top row
                                     #########################################
if panoheadPhiMax >= optimalZenith \
    and truePhiMax > 90 - (trueHfov / 2 * (1 - max(Voverlap, Hoverlap))) :
    zenith= "yes"
    phiMax= phiMax - (min(trueHfov, trueVfov) / 2 * (1 - max(Voverlap, Hoverlap)))
    totalImgNb= totalImgNb + 1
else :
    zenith= "no"                        # no overlap with zenith hole is wanted
    phiMax= phiMax - (trueVfov * Voverlap / 2)                                
    print("# phiMax moved downward by", repr(trueVfov * Voverlap / 2)[0:5],
          "deg -----------------------> 0 % overlap with zenith hole")

if (panoheadPhiMin <= optimalNadir) \
    and abs(truePhiMin) > 90 - (trueVfov / 2 * (1 - max(Voverlap, Hoverlap))) :
    nadir=  "yes"
    phiMin= phiMin + (min(trueHfov, trueVfov) / 2 * (1 - max(Voverlap, Hoverlap)))
    totalImgNb= totalImgNb + 1
else :
    nadir= "no"                          # no overlap with nadir hole is wanted
    phiMin= phiMin + (trueVfov * Voverlap / 2) 
    print("# phiMin moved upward by", repr(trueVfov * Voverlap / 2)[0:5],
          "deg -------------------------> 0 % overlap with nadir hole")
  
Vfov=   trueVfov  * (1 - Voverlap)
Hfov=   trueHfov  * (1 - Hoverlap)
Vside=  2 * f * tan(radians(Vfov / 2))
Hside=  2 * f * tan(radians(Hfov / 2))

linV=   1 - Vside / trueVside # Out of disc part of zenith and/or nadir 
linH=   1 - Hside / trueHside # not taken into account tough it should be!
if int(linV) != int(linH) :
    print("# Linear vertical overlap   --> ", repr(100 * (linV))[0:5],
          "% of source pixels")
    print("# Linear horizontal overlap --> ", repr(100 * (linH))[0:5],
          "% of source pixels")
else :
    print("# Linear overlap --> ", repr(100 * (linV))[0:5],
          "% of source pixels")

anglV=  1 - Vfov / trueVfov # idem about out of disc part of zenith, nadir
anglH=  1 - Hfov / trueHfov
if int(100 * (anglV + anglH)) != int(100 * (linV + linH)) :
    if int(anglV) != int(anglH) :
        print("# Angular vertical overlap   --> ", repr(100 * anglV)[0:5], "%")
        print("# Angular horizontal overlap --> ", repr(100 * anglH)[0:5], "%")
    else :
        print("# Angular overlap --> ", repr(100 * anglV)[0:5], "%")

rowNb= (phiMax - phiMin) // Vfov
if rowNb * Vfov < phiMax - phiMin :
    rowNb= rowNb + 1
dPitch= (phiMax - phiMin) / rowNb               

sigmaPhi= rowNb * trueVfov
if nadir == "yes" :
    sigmaPhi= sigmaPhi + (trueVfov + trueHfov )/ 2 / 2 # sic!
if zenith == "yes" :
    sigmaPhi= sigmaPhi +  + (trueVfov + trueHfov )/ 2 / 2  # sic!
VangularExcess= (sigmaPhi - (truePhiMax - truePhiMin)) / sigmaPhi
    
##########----------##########----------##########----------##########----------

toPrint= "\n" + "Number of rows: " + repr(int(rowNb))
if zenith == "yes" :
    toPrint= toPrint + " + zenith"
if nadir == "yes" :
    toPrint= toPrint + " + nadir"
print (toPrint + "\n")

print("# Lens vertical angles :    ",
      repr(trueVfov * Voverlap / 2)[0:4], "+", repr(Vfov)[0:4], "+",
      repr(trueVfov * Voverlap / 2)[0:4], "=",repr(trueVfov)[0:4] + " deg")
print("# Lens horizontal angles :  ",
      repr(trueHfov * Hoverlap / 2)[0:4], "+", repr(Hfov)[0:4], "+",
      repr(trueHfov * Hoverlap / 2)[0:4], "=", repr(trueHfov)[0:4] + " deg")
if zenith == "yes" or nadir == "yes" :
    if Hfov < Vfov :
        print("# Zenith and Nadir angles:  ", repr(Hfov / 2)[0:4], "+",
              repr(trueHfov * Hoverlap / 2)[0:4], "=",
              repr(trueHfov / 2)[0:4] + "deg")
    else :
        print("# Zenith and Nadir angles:  ", repr(Vfov / 2)[0:4], "+",
              repr(trueVfov * Voverlap / 2)[0:4], "=",
              repr(trueVfov / 2)[0:4] + " deg")


print("# Modified phiMax = " + repr(phiMax)[0:4] + " deg")
print("# Modified phiMin = " + repr(phiMin)[0:5] + " deg")
print("# Row to row delta Pitch = " +
      repr(dPitch)[0:4] + " deg -----------------------> " + 
      repr(int(100 * VangularExcess)) +
      " % vertical angular excess (total angle: " +
      repr(int(sigmaPhi)) + " deg)")
print("#############################")

##########----------##########----------##########----------##########----------

if nadir== "yes" :
    if trueVfov > trueHfov :
        print("\nNadir shoot at " + repr(optimalNadir)[0:5] + " deg\n")
    else :
        print("\nNadir shoot at -90 deg\n")
    pitchList=  [repr(optimalNadir)[0:5]]           ## 1 ##
    imgNbList=  [int(1)]                            ## 2 ##
    imgYawList= ["0.0"]                             ## 3 ##
else :
    pitchList=  []                                  ## 1 ##
    imgNbList=  []                                  ## 2 ##
    imgYawList= []                                  ## 3 ##


if rowNb != 1 :
    prevTop= phiMin
else :
    prevTop= (phiMin + phiMax) / 2 - Vfov / 2
i= 0

while i < rowNb :                              # begin rows loop <----------------------------- 
    bottom= prevTop
    top= prevTop + dPitch
    pitch= (bottom + top) / 2

    if pitch < 0 :
        pitchList= pitchList + [repr(pitch)[0:5]]    ## 1 ##
    else :
        pitchList= pitchList + [repr(pitch)[0:4]]    ## 1 ##

    if abs(bottom) < abs(top) :        # the smaller the distance 
        largeCirclePitch= abs(bottom)  # between a circle and the equator
        smallCirclePitch= abs(top)     # ... the larger this circle is
    else:
        largeCirclePitch= abs(top)
        smallCirclePitch= abs(bottom)

    largePitch=  largeCirclePitch                                 
    midlePitch=  (largeCirclePitch + smallCirclePitch) / 2        
    smallPitch=  smallCirclePitch                                 

# midlePitch (degrees, unsigned) points to the circle where images
# sides must be adjacent

    MoHSD= f / cos(radians(dPitch / 2)) # MoHSD: Middle of Horizontal Sides 
                                        # Distance from sphere center
    circleRadius= MoHSD * cos (radians(midlePitch))
    largeRadius=  MoHSD * cos (radians(largePitch))
    smallRadius=  MoHSD * cos (radians(smallPitch))
                                                                     
    # angle under which is seen an image (decreased for overlap) in the:
    unadjdYaw= degrees(2 * (atan(Hside / 2 / circleRadius)))   # middle circle horizontal plane
    largedYaw= degrees(2 * (atan(trueHside / 2 / largeRadius)))# largest circle horizontal plane
    smalldYaw= degrees(2 * (atan(trueHside / 2 / smallRadius)))# smallest circle horizontal plane


    imgNb= 360 // unadjdYaw
    if imgNb * unadjdYaw < 360 :
        imgNb = imgNb + 1
    dYaw = 360 / imgNb
    if largedYaw < dYaw :
        imgNb= 360 // largedYaw
        if imgNb * largedYaw < 360 :
            imgNb = imgNb + 1
        dYaw = 360 / imgNb
        selected= "# based on LARGE circle, dYaw: " + repr(dYaw)[0:5] + " deg"
    else :
        selected= "# based on MIDLE circle, dYaw: " + repr(dYaw)[0:5] + " deg"

    
    imgNbList= imgNbList + [int(imgNb)]                ## 2 ##
    HexcessContiguous= (imgNb * unadjdYaw - 360) / 360

    toPrint = "\nrow " + repr(i + 1).ljust(2) + " of " + repr(int(rowNb)) \
            + ", pitch: " + repr(int(pitch)) \
            + " deg      # top overlap at " + repr(int(round(top))) \
            + " deg, bottom overlap at " + repr(int(round(bottom))) + " deg" 
    print (toPrint)
    print("                                " + selected)

    if abs(round(top)) == abs(round(bottom)) :
        toPrint= "                                # large circle is any -----> "
    elif largeCirclePitch == abs(bottom) :
        toPrint= "                                # large circle at bottom --> "
    else:
        toPrint= "                                # large circle at top -----> "

    HexcessLarge= (largedYaw - dYaw) / largedYaw
    toPrint= toPrint + repr(int(100 * HexcessLarge)) + " % overlap"
    print (toPrint)

    toPrint=     "             deg                # small circle ------------> "
    HexcessSmall= (smalldYaw - dYaw) / smalldYaw
    toPrint= toPrint + repr(int(100 * HexcessSmall)) + " % overlap"
    print (toPrint)

    dYaw = 360 / imgNb
    j= 0                                  # begin images loop <----------------- 
    while j < imgNb :
        imgYawList= imgYawList + [repr(j * dYaw)[0:5]]  ## 3 ##
        totalImgNb= totalImgNb + 1

        print ("    ", repr(j + 1).ljust(2), "yaw:", int(j * dYaw))
        
        j= j + 1                          # end   images loop <-----------------
    i= i + 1                             
    prevTop= top                          # end   rows loop <-------------------

##########----------##########----------##########----------##########----------

if zenith == "yes" :
    if trueVfov > trueHfov :
        print("\nZenith shoot at " + repr(optimalZenith)[0:4] + " deg")
    else :
        print("\nZenith shoot at 90 deg")
    pitchList= pitchList +  [repr(optimalZenith)[0:4]]   ## 1 ##
    imgNbList= imgNbList + [int(1)]                      ## 2 ##
    imgYawList= imgYawList + ["0.0"]                     ## 3 ##

print("\nTotal number of images: " + repr(totalImgNb) + "\n")

##########----------##########----------##########----------##########----------
xml_file = open(repr(f)[0:5]+ "mm sphere_preset.xml", 'w')
header = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<papywizard>\n    <preset name=\""+repr(f)[0:5]+ "mm sphere\">\n        <tooltip>\n"
xml_file.write(header)
xml_file.write("Total number of images: " + repr(totalImgNb) + "\n")
xml_file.write("Camera:\n")
xml_file.write("    Sensor size         "+ repr(trueHside)[0:4] + " x " + repr(trueVside)+"mm\n")
xml_file.write("    Focal length        "+ repr(f)[0:5]+ "mm\n")
xml_file.write("    Lens vertical FOV   "+ str(round(degrees(2 * atan (trueVside / 2 / f))))+ "deg\n")
xml_file.write("    Lens horizontal FOV "+ str(round(degrees(2 * atan (trueHside / 2 / f))))+ "deg\n")

xml_file.write("\nPanorama:\n")
xml_file.write("    Top at              "+ repr(phiMax)[0:5]+ "deg\n")
xml_file.write("    Bottom at           "+ repr(phiMin)[0:5]+ "deg\n")
xml_file.write("    vertical FOV        "+ repr(phiMax - phiMin)[0:5]+ "deg\n")
xml_file.write("    Vertical overlap    "+ repr(int(100 * Voverlap))+" %\n")
xml_file.write("    Horizontal overlap  "+ repr(int(100 * Hoverlap))+" %\n")
header2 = "        </tooltip>\n        <shoot>\n"
xml_file.write(header2)

print("#############################\n")
k= 0
for i in range(len(pitchList)) :
    pitchString= 'pitch="' + pitchList[i] + '" yaw="'
    if pitchString[7] == '-' and pitchString[9] == '.' :
        pitchString= pitchString[0:10]+pitchString[11:]
    if i%2!=0:
        for j in range(imgNbList[i]) :
            yawString= imgYawList[k] + '"'; k= k + 1;
            print("<pict " + pitchString + yawString +"/>")
            xml_file.write("           <pict " + pitchString + yawString +"/>\n")
    else:
        for j in range(imgNbList[i]) :
            yawString= imgYawList[k+imgNbList[i]-2*j-1] + '"'; k= k + 1;
            print("<pict " + pitchString + yawString +"/>")
            xml_file.write("           <pict " + pitchString + yawString +"/>\n")

    print("")
    xml_file.write("\n")
footer = "         </shoot>\n    </preset>\n\n</papywizard>"
xml_file.write(footer)
xml_file.close()    
print("#############################")

print("Papywizard preset XML saved as:  "+repr(f)[0:5]+ "mm sphere_preset.xml")
#print("the following is buggy...")
#totalArea= totalImgNb * trueVside * trueHside
#imgDiag = sqrt((trueVside)**2 + (trueHside)**2)
#r = sqrt(f**2 + (imgDiag / 2)**2)
#wrapingCylinderHeight = r * (  sin(radians(abs(truePhiMax)))
#                             + sin(radians(abs(truePhiMin))))
#wrapingCylinderArea= 2 * pi * r * wrapingCylinderHeight
######################
